from typing import List, Literal, Optional, Tuple

from grai_client.endpoints.client import BaseClient
from grai_client.update import update
from grai_schemas.base import Node, Edge

from adapters import adapt_to_client
from loader import MetabaseConnector


def get_nodes_and_edges(
    connector: MetabaseConnector, version: Literal["v1"]
) -> Tuple[List[Node], List[Edge]]:
    """

    Args:
        connector (MetabaseConnector):
        version (Literal["v1"]):

    Returns:

    Raises:

    """
    nodes = connector.get_nodes()
    edges = connector.get_edges()

    nodes = adapt_to_client(nodes, version)
    edges = adapt_to_client(edges, version)
    return nodes, edges


def update_server(
    client: BaseClient,
    namespaces: Optional = None,
    default_namespace: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    endpoint: Optional[str] = None,
):
    """

    Args:
        password: Optional[str]
        username: Optional[str]
        client (BaseClient):
        namespaces (Optional[NamespaceTypes], optional):  (Default value = None)
        default_namespace (Optional[str], optional):  (Default value = None)
        endpoint (Optional[str], optional):  (Default value = None)

    Returns:

    Raises:

    """
    kwargs = {
        "namespaces": namespaces,
        "default_namespace": default_namespace,
        "username": username,
        "password": password,
        "endpoint": endpoint,
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    conn = MetabaseConnector(**kwargs)
    nodes, edges = get_nodes_and_edges(conn, "v1")

    update(client, nodes)
    update(client, edges)



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel


class Question(BaseModel):
    id: int
    name: str
    creator: Dict[str, Any]
    database_id: int
    namespace: str

    @property
    def full_name(self):
        return f"{self.name}.{self.id}"


class Table(BaseModel):
    id: int
    db_id: int
    name: str
    display_name: str
    schema_name: str
    db: Dict[str, Any]
    namespace: str

    @property
    def full_name(self):
        return f"{self.schema_name}.{self.name}"


NodeTypes = Union[Question, Table]


class Edge(BaseModel):
    source: NodeTypes
    destination: NodeTypes
    definition: Optional[str]
    metadata: Optional[Dict] = None

    def __hash__(self):
        return hash((self.source, self.destination))



import importlib
from functools import wraps
from typing import Callable, TypeVar

from grai_source_metabase.models import Question, Edge, Table

T = TypeVar("T")

has_faker = importlib.util.find_spec("faker") is not None
if has_faker:
    from faker import Faker

    fake = Faker()


def faker_dep_wrapper(fn: Callable[..., T]):
    """

    Args:
        fn (Callable[..., T]):

    Returns:

    Raises:

    """

    @wraps(fn)
    def inner(*args, **kwargs) -> T:
        """

        Args:
            *args:
            **kwargs:

        Returns:

        Raises:

        """
        if not has_faker:
            raise ModuleNotFoundError(
                "Mock testing tools require `faker`. Try running `pip install faker`"
            )
        return fn(*args, **kwargs)

    return inner


class MockMetabaseObjects:
    """ """

    @staticmethod
    @faker_dep_wrapper
    def mock_question():
        """ """
        return Question(
            name=fake.name(),
            namespace=fake.color_name(),
            description=fake.sentence(),
            database_id=fake.pyint(1, 15),
            table_id=fake.pyint(1, 10000),
            public_uuid=fake.uuid4(),
            id=fake.pyint(1, 10000),
            result_metadata=[
                {
                    "display_name": "display_name",
                    "name": "display_name",
                    "base_type": "type/Text",
                    "effective_type": "type/Text",
                    "fingerprint": {
                        "global": {"distinct-count": 1, "nil%": 0, "sample": "sample"}
                    },
                }
            ],
            creator={
                "id": fake.pyint(1, 2000),
                "first_name": fake.first_name(),
                "last_name": fake.last_name(),
                "email": fake.email(),
            },
            collection={
                "id": fake.pyint(1, 2000),
                "name": fake.name(),
                "color": fake.color_name(),
                "slug": fake.slug(),
                "description": fake.sentence(),
            },
        )

    @staticmethod
    @faker_dep_wrapper
    def mock_table():
        """ """
        return Table(
            name=fake.name(),
            namespace=fake.color_name(),
            id=fake.pyint(1, 10000),
            db_id=fake.pyint(1, 15),
            schema_name=fake.emoji(),
            description=fake.sentence(),
            display_name=fake.name(),
            entity_type="entity/UserTable",
            db={"id": fake.pyint(1, 15), "name": fake.name(), "engine": "postgres"},
        )

    @classmethod
    @faker_dep_wrapper
    def mock_edge(cls, type: str = "tq"):
        """

        Args:
            type (str, optional):  (Default value = "tq")

        Returns:

        Raises:

        """

        def mock_node(type: str):
            """

            Args:
                type (str):

            Returns:

            Raises:

            """
            if type == "t":
                return cls.mock_table()
            elif type == "q":
                return cls.mock_question()

        source = mock_node(type[0])
        destination = mock_node(type[1])
        return Edge(source=source, destination=destination, constraint_type="bt")



import json
from itertools import chain
from typing import Callable, Dict, List, Optional, Union

import requests
from grai_source_metabase.models import Edge, NodeTypes, Question, Table
from pydantic import BaseSettings, SecretStr, validator
from requests.exceptions import ConnectionError
from retrying import retry


class MetabaseConfig(BaseSettings):
    endpoint: str
    username: SecretStr
    password: SecretStr

    @validator("endpoint")
    def validate_endpoint(cls, endpoint: str):
        endpoint = endpoint.rstrip("/")
        return endpoint

    class Config:
        env_prefix = "grai_metabase_"
        env_file = ".env"


class MetabaseAPI:
    """
    Wrapper class for interacting with the Metabase API.

    Args:
        username (str, optional): Metabase username. Defaults to None or returns value from environment variable.
        password (str, optional): Metabase password. Defaults to None or returns value from environment variable.
        endpoint (str, optional): Metabase API endpoint URL. Defaults to None.

    """

    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        endpoint: Optional[str] = None,
    ):
        passthrough_kwargs = {
            "username": username,
            "password": password,
            "endpoint": endpoint,
        }

        self.config = MetabaseConfig(
            **{k: v for k, v in passthrough_kwargs.items() if v is not None}
        )
        self.api_endpoint = f"{self.config.endpoint}/api"

        self.session = requests.Session()
        self.session.headers.update({"Content-Type": "application/json"})
        self.session.headers.update(self.authenticate())

    @retry(stop_max_attempt_number=3, wait_fixed=5000)
    def authenticate(self) -> Dict[str, str]:
        """
        Authenticates the user and sets the session headers.

        Args:
            request (callable): The HTTP request method to use for authentication.
            url (str): The URL for the authentication endpoint.

        Raises:
            ConnectionError: If there is an error connecting to the endpoint.

        """
        try:
            response = requests.post(
                url=f"{self.api_endpoint}/session",
                json={
                    "username": self.config.username.get_secret_value(),
                    "password": self.config.password.get_secret_value(),
                },
            )

            response.raise_for_status()
            return {"X-Metabase-Session": response.json()["id"]}

        except ConnectionError as ce:
            raise ce

    @staticmethod
    def make_request(request: Callable[..., requests.Response], url: str):
        """
        Makes an authenticated API request and returns the JSON response.

        Args:
            request (callable): The HTTP request method to use for the API call.
            url (str): The URL for the API endpoint.

        Returns:
            dict: The JSON response from the API.

        Raises:
            AssertionError: If the response status code is not 200.

        """
        response = request(url=url)
        assert response.status_code == 200
        return response.json()

    def get_questions(self):
        """
        Retrieves the list of questions from the Metabase API.

        Returns:
            dict: The JSON response containing the list of questions.

        """
        url = f"{self.api_endpoint}/card"
        return self.make_request(self.session.get, url)

    def get_tables(self):
        """
        Retrieves the list of tables from the Metabase API.

        Returns:
            dict: The JSON response containing the list of tables.

        """
        url = f"{self.api_endpoint}/table"
        return self.make_request(self.session.get, url)

    def get_dbs(self):
        """
        Retrieves the list of databases from the Metabase API.

        Returns:
            dict: The JSON response containing the list of databases.

        """
        url = f"{self.api_endpoint}/database"
        return self.make_request(self.session.get, url)

    def get_collections(self):
        """
        Retrieves the list of collections from the Metabase API.

        Returns:
            dict: The JSON response containing the list of collections.

        """

        pass


def build_namespace_map(
    dbs: Dict, namespace_map: Union[str, Dict, None], metabase_namespace: str
) -> Dict:
    if isinstance(namespace_map, str):
        namespace_map = json.loads(namespace_map)
    elif namespace_map is None:
        namespace_map = {}

    for k, v in dbs.items():
        db_id_ns = f"{metabase_namespace}.{k}.{v['name']}"
        namespace_map.setdefault(k, db_id_ns)

    return namespace_map


class MetabaseConnector(MetabaseAPI):
    """
    Connector class for interacting with Metabase API and building lineage information.

    Args:
        namespaces (Dict, optional): A mapping of database IDs to their corresponding namespace names. Defaults to None.
        default_namespace (str, optional): The default namespace to be used when a table or question does not have a specific namespace. Defaults to None.
        *args: Additional positional arguments to be passed to the base class constructor.
        **kwargs: Additional keyword arguments to be passed to the base class constructor.

    Attributes:
        default_namespace (str): The default namespace to be used.
        tables (List[Dict]): The list of tables retrieved from the Metabase API.
        tables_map (Dict[int, Dict]): A mapping of table IDs to their corresponding table dictionaries.
        dbs_map (Dict[int, Dict]): A mapping of database IDs to their corresponding database dictionaries.
        questions_map (Dict[int, Dict]): A mapping of question IDs to their corresponding question dictionaries.
        namespace_map (Dict[int, str]): A mapping of database IDs to their corresponding namespace names.
        question_table_map (Dict[int, int]): A mapping of question IDs to their corresponding table IDs.
        table_db_map (Dict[int, int]): A mapping of table IDs to their corresponding database IDs.
        question_db_map (Dict[int, int]): A mapping of question IDs to their corresponding database IDs.

    """

    def __init__(
        self,
        metabase_namespace: str,
        namespaces: Optional[Dict] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._lineage_ready = False
        self.metabase_namespace = metabase_namespace

        # This line creates a list of tables by modifying each table dictionary obtained from the get_tables() method.
        # It replaces the "schema" key with a new key "schema_name" while preserving the other key-value pairs.
        # this is because the "schema" key is a reserved keyword in the pydantic.
        self.tables = [
            {**table, "schema_name": table.pop("schema")} for table in self.get_tables()
        ]
        self.tables_map = {
            table["id"]: table for table in self.tables if table["active"]
        }
        self.dbs_map = {db["id"]: db for db in self.get_dbs()["data"]}

        self.questions_map = {
            question["id"]: question
            for question in self.get_questions()
            if question["archived"] is False
        }
        self.namespace_map = build_namespace_map(
            self.dbs_map, namespaces, self.metabase_namespace
        )

        self.question_table_map = {}
        self.table_db_map = {}
        # self.question_db_map = {}

    def build_lineage(self):
        """
        Builds the lineage information by updating the mappings and maps.

        This method updates the `tables_map`, `dbs_map`, `questions_map`, `question_table_map`, `table_db_map`,
        and `question_db_map` attributes.

        """

        self.question_table_map = {
            question["id"]: question["table_id"]
            for question in self.questions_map.values()
            if question["table_id"]
            and self.tables_map.get(question["table_id"]) is not None
        }

        self.table_db_map = {
            table["id"]: table["db_id"]
            for table in self.tables
            if table["id"] is not None and table["db_id"] is not None
        }

        # for question, table in self.question_table_map.items():
        #     self.question_db_map[question] = self.table_db_map[table]

        self._lineage_ready = True

    def get_nodes(self) -> List[NodeTypes]:
        """
        Retrieves the list of nodes representing tables and questions.

        Returns:
            List[NodeTypes]: The list of nodes.

        """

        if not self._lineage_ready:
            self.build_lineage()

        for question in self.questions_map.values():
            question["namespace"] = self.metabase_namespace

        for table in self.tables_map.values():
            table["namespace"] = self.namespace_map[self.table_db_map[table["id"]]]

        verified_questions = [
            Question(**question) for question in self.questions_map.values()
        ]
        verified_tables = [Table(**table) for table in self.tables_map.values()]
        nodes = chain(verified_questions, verified_tables)

        return list(nodes)

    def get_edges(self) -> List[Edge]:
        """
        Retrieves the list of edges representing the relationships between questions and tables.

        Returns:
            List[Edge]: The list of edges.

        """

        if not self._lineage_ready:
            self.build_lineage()

        edges = (
            Edge(
                source=Question(**self.questions_map[question]),
                destination=Table(**self.tables_map[table]),
                namespace=self.questions_map[question]["namespace"],
            )
            for question, table in self.question_table_map.items()
        )
        return list(edges)



from typing import Any, Dict, List, Literal, Sequence, Union

from grai_schemas import config as base_config
from grai_schemas.schema import Schema
from grai_schemas.v1 import EdgeV1, NodeV1
from grai_schemas.v1.metadata.edges import EdgeMetadataTypeLabels, GenericEdgeMetadataV1
from grai_schemas.v1.metadata.nodes import (
    GenericNodeMetadataV1,
    NodeMetadataTypeLabels,
    TableMetadata,
)
from grai_source_metabase.models import Edge, NodeTypes, Question, Table
from grai_source_metabase.package_definitions import config
from multimethod import multimethod
from pydantic import BaseModel


@multimethod
def build_grai_metadata(current: Any, desired: Any) -> None:
    """
    Build grai metadata for a given object.

    Args:
        current (Any): The object to build grai metadata from.
        desired (Any): The desired format of the metadata.

    Returns:
        None: grai metadata object.

    Raises:
        NotImplementedError: If no adapter is available between the `current` and `desired` types.

    """

    raise NotImplementedError(
        f"No adapter between {type(current)} and {type(desired)} for value {current}"
    )


@build_grai_metadata.register
def build_grai_metadata_from_table(
    current: Table, version: Literal["v1"] = "v1"
) -> TableMetadata:
    """
    Build grai metadata for a Table object.

    Args:
        current (Table): The Table object to build grai metadata from.
        version (Literal["v1"], optional): The version of grai metadata to build. Defaults to "v1".

    Returns:
        TableMetadata: grai metadata object for the Table.

    Raises:
        None.

    """

    data = {
        "version": version,
        "node_type": NodeMetadataTypeLabels.table.value,
        "node_attributes": {},
        "tags": [config.metadata_id],
    }

    return TableMetadata(**data)


@build_grai_metadata.register
def build_grai_metadata_from_question(
    current: Question, version: Literal["v1"] = "v1"
) -> GenericNodeMetadataV1:
    """
    Build grai metadata for a Question object.

    Args:
        current (Question): The Question object to build grai metadata from.
        version (Literal["v1"], optional): The version of grai metadata to build. Defaults to "v1".

    Returns:
        GenericNodeMetadataV1: grai metadata object for the Question.

    Raises:
        None.

    """

    data = {
        "version": version,
        "node_type": NodeMetadataTypeLabels.generic.value,
        "node_attributes": {},
        "tags": [config.metadata_id],
    }

    return GenericNodeMetadataV1(**data)


@build_grai_metadata.register
def build_grai_metadata_from_edge(
    current: Edge, version: Literal["v1"] = "v1"
) -> GenericEdgeMetadataV1:
    """
    Build grai metadata for an Edge object.

    Args:
        current (Edge): The Edge object to build grai metadata from.
        version (Literal["v1"], optional): The version of grai metadata to build. Defaults to "v1".

    Returns:
        GenericEdgeMetadataV1: grai metadata object for the Edge.

    Raises:
        None.

    """

    data = {
        "version": version,
        "edge_type": EdgeMetadataTypeLabels.generic.value,
        "tags": [config.metadata_id],
    }

    return GenericEdgeMetadataV1(**data)


@multimethod
def build_app_metadata(current: Any, desired: Any) -> None:
    """
    Build application-specific metadata for a given object.

    Args:
        current (Any): The object to build application-specific metadata from.
        desired (Any): The desired format of the metadata.

    Returns:
        None: Application-specific metadata object.

    Raises:
        NotImplementedError: If no adapter is available between the `current` and `desired` types.

    """

    raise NotImplementedError(
        f"No adapter between {type(current)} and {type(desired)} for value {current}"
    )


@build_app_metadata.register
def build_metadata_from_table(
    current: BaseModel, version: Literal["v1"] = "v1"
) -> Dict:
    """
    Build application-specific metadata for a Table object.

    Args:
        current (Table): The Table object to build application-specific metadata from.
        version (Literal["v1"], optional): The version of the metadata to build. Defaults to "v1".

    Returns:
        Dict: Application-specific metadata object for the Table.

    Raises:
        None.

    """

    return {"raw_object": current.dict()}


@build_app_metadata.register
def build_metadata_from_edge(current: Edge, version: Literal["v1"] = "v1") -> Dict:
    """
    Build application-specific metadata for an Edge object.

    Args:
        current (Edge): The Edge object to build application-specific metadata from.
        version (Literal["v1"], optional): The version of the metadata to build. Defaults to "v1".

    Returns:
        Dict: Application-specific metadata object for the Edge.

    Raises:
        None.

    """

    data = {
        "definition": current.definition,
        "metadata": current.metadata if current.metadata is not None else {},
        "raw_object": current.dict(),
    }

    return data


def build_metadata(obj, version):
    """
    Build metadata for a given object.

    Args:
        obj: The object to build metadata from.
        version: The version of the metadata to build.

    Returns:
        Dict: Metadata object containing both grai and application-specific metadata.

    Raises:
        None.

    """
    return {
        base_config.metadata_id: build_grai_metadata(obj, version),
        config.metadata_id: build_app_metadata(obj, version),
    }


@multimethod
def adapt_to_client(current: Any, desired: Any):
    """
    Adapt a given object to the desired client format.

    Args:
        current (Any): The object to adapt.
        desired (Any): The desired format to adapt to.

    Returns:
        None: Adapted object in the desired format.

    Raises:
        NotImplementedError: If no adapter is available between the `current` and `desired` types.

    """

    raise NotImplementedError(f"No adapter between {type(current)} and {type(desired)}")


@adapt_to_client.register
def adapt_table_to_client(current: Table, version: Literal["v1"] = "v1") -> NodeV1:
    """
    Adapt a Table object to the desired client format.

    Args:
        current (Table): The Table object to adapt.
        version (Literal["v1"], optional): The version of the client format to adapt to. Defaults to "v1".

    Returns:
        NodeV1: Adapted Table object in the desired client format.

    Raises:
        None.

    """

    spec_dict = {
        "name": current.full_name,
        "namespace": current.namespace,
        "display_name": current.name,
        "data_source": config.integration_name,
        "metadata": build_metadata(current, version),
    }

    return Schema.to_model(spec_dict, version=version, typing_type="Node")


@adapt_to_client.register
def adapt_question_to_client(
    current: Question, version: Literal["v1"] = "v1"
) -> NodeV1:
    """
    Adapt a Question object to the desired client format.

    Args:
        current (Question): The Question object to adapt.
        version (Literal["v1"], optional): The version of the client format to adapt to. Defaults to "v1".

    Returns:
        NodeV1: Adapted Question object in the desired client format.

    Raises:
        None.

    """

    spec_dict = {
        "name": current.full_name,
        "namespace": current.namespace,
        "display_name": current.name,
        "data_source": config.integration_name,
        "metadata": build_metadata(current, version),
    }

    return Schema.to_model(spec_dict, version=version, typing_type="Node")


def make_name(node1: NodeTypes, node2: NodeTypes) -> str:
    """
    Creates a name for an edge based on the given nodes.

    Args:
        node1 (NodeTypes)
        node2 (NodeTypes)

    Returns:
        str: The name of the edge.

    Raises:
        None.
    """

    node1_name = f"{node1.namespace}:{node1.full_name}"
    node2_name = f"{node2.namespace}:{node2.full_name}"

    return f"{node1_name} -> {node2_name}"


@adapt_to_client.register
def adapt_edge_to_client(current: Edge, version: Literal["v1"] = "v1") -> EdgeV1:
    """
    Adapt an Edge object to the desired client format.

    Args:
        current (Edge): The Edge object to adapt.
        version (Literal["v1"], optional): The version of the client format to adapt to. Defaults to "v1".

    Returns:
        EdgeV1: Adapted Edge object in the desired client format.

    Raises:
        None.

    """

    spec_dict = {
        "data_source": config.integration_name,
        "name": make_name(current.source, current.destination),
        "namespace": current.source.namespace,
        "source": {
            "name": current.source.full_name,
            "namespace": current.source.namespace,
        },
        "destination": {
            "name": current.destination.full_name,
            "namespace": current.destination.namespace,
        },
        "metadata": build_metadata(current, version),
    }

    return Schema.to_model(spec_dict, version=version, typing_type="Edge")


@adapt_to_client.register
def adapt_seq_to_client(
    objs: Sequence, version: Literal["v1"]
) -> List[Union[NodeV1, EdgeV1]]:
    """
    Adapt a sequence of objects to the desired client format.

    Args:
        objs (Sequence): The sequence of objects to adapt.
        version (Literal["v1"]): The version of the client format to adapt to.

    Returns:
        List[Union[NodeV1, EdgeV1]]: Adapted sequence of objects in the desired client format.

    Raises:
        None.

    """
    entities = [adapt_to_client(item, version) for item in objs]
    entities = {entity for entity in entities}
    return list(entities)


@adapt_to_client.register
def adapt_list_to_client(
    objs: List, version: Literal["v1"]
) -> List[Union[NodeV1, EdgeV1]]:
    """
    Adapt a list of objects to the desired client format.

    Args:
        objs (List): The list of objects to adapt.
        version (Literal["v1"]): The version of the client format to adapt to.

    Returns:
        List[Union[NodeV1, EdgeV1]]: Adapted list of objects in the desired client format.

    Raises:
        None.

    """

    entities = [adapt_to_client(item, version) for item in objs]
    entities = {entity for entity in entities}
    return list(entities)
